@@ -1,31 +1,42 @@
 from distutils.version import LooseVersion
+import functools
+import django
 from django.conf import settings
 from django.contrib.admin.options import ModelAdmin, csrf_protect_m, InlineModelAdmin
-from django.contrib.admin.util import (flatten_fieldsets, unquote, 
-    get_deleted_objects)
+if django.VERSION >= (1, 7):
+    from django.contrib.admin.utils import (flatten_fieldsets, unquote,
+        get_deleted_objects)
+else:
+    from django.contrib.admin.util import (flatten_fieldsets, unquote,
+        get_deleted_objects)
+from django.contrib.contenttypes.models import ContentType
 from django.core.exceptions import ValidationError, PermissionDenied
 from django.core.urlresolvers import reverse
 from django.db import router, transaction
-from django.forms.formsets import formset_factory
-from django.forms.models import ModelForm, BaseModelFormSet, BaseInlineFormSet, model_to_dict
-from django.forms.util import ErrorList
+from django.forms.models import model_to_dict
+if django.VERSION >= (1, 7):
+    from django.forms.utils import ErrorList
+else:
+    from django.forms.util import ErrorList
 from django.http import Http404, HttpResponseRedirect
 from django.shortcuts import render_to_response
 from django.template import TemplateDoesNotExist
 from django.template.context import RequestContext
 from django.template.loader import find_template
-from django.utils.encoding import iri_to_uri, force_unicode
+from django.utils.encoding import iri_to_uri
 from django.utils.functional import curry
 from django.utils.translation import ugettext_lazy as _, get_language
 from functools import update_wrapper
+from hvad.compat.force_unicode import force_unicode
+from hvad.compat.urls import urlencode
 from hvad.forms import TranslatableModelForm, translatable_inlineformset_factory, translatable_modelform_factory
-import django
-import urllib
 from hvad.utils import get_cached_translation, get_translation
 from hvad.manager import FALLBACK_LANGUAGES
 
 
 NEW_GET_DELETE_OBJECTS = LooseVersion(django.get_version()) >= LooseVersion('1.3')
+atomic = (transaction.atomic if django.VERSION >= (1, 6) else
+          transaction.commit_on_success)
 
 
 def get_language_name(language_code):
@@ -54,7 +65,7 @@ class InlineModelForm(TranslatableModelForm):
                 # Dirty hack that swaps the id from the translation id, to the master id
                 # This is necessary, because we in this case get the untranslated instance,
                 # and thereafter get the correct translation on save.
-                if object_data.has_key("id"):
+                if "id" in object_data:
                     object_data["id"] = trans.master.id
         if initial is not None:
             object_data.update(initial)
@@ -81,11 +92,11 @@ class TranslatableModelAdminMixin(object):
             languages = []
             current_language = get_language()
             for language in obj.get_available_languages():
+                entry = '<a href="%s">%s</a>' % (self.get_url(obj, lang=language), language)
                 if language == current_language:
-                    languages.append(u'<strong>%s</strong>' % language)
-                else:
-                    languages.append(language)
-            return u' '.join(languages)
+                    entry = u'<strong>%s</strong>' % entry
+                languages.append(entry)
+            return u', '.join(languages)
         else:
             return ''
     all_translations.allow_tags = True
@@ -103,9 +114,7 @@ class TranslatableModelAdminMixin(object):
         language = self._language(request)
         for key, name in settings.LANGUAGES:
             get.update({'language': key})
-            url = '%s://%s%s?%s' % (request.is_secure() and 'https' or 'http',
-                                    request.get_host(), request.path,
-                                    urllib.urlencode(get))
+            url = '%s?%s' % (request.path, urlencode(get))
             if language == key:
                 status = 'current'
             elif key in available_languages:
@@ -126,8 +135,25 @@ class TranslatableAdmin(ModelAdmin, TranslatableModelAdminMixin):
     
     deletion_not_allowed_template = 'admin/hvad/deletion_not_allowed.html'
     
+    def __init__(self, *args, **kwargs):
+        super(TranslatableAdmin, self).__init__(*args, **kwargs)
+        self.reverse = functools.partial(reverse, current_app=self.admin_site.name)
+
+
+    def get_url(self, obj, lang=None, get={}):
+        ct = ContentType.objects.get_for_model(self.model)
+        info = ct.app_label, ct.model
+        if lang:
+            get.update({self.query_language_key: lang})
+        url = '%s?%s' % (self.reverse('admin:%s_%s_change' % info, args=(obj.id,)), urlencode(get))
+        return url
+
+
     def get_urls(self):
-        from django.conf.urls.defaults import patterns, url
+        try:
+            from django.conf.urls import patterns, url
+        except ImportError:
+            from django.conf.urls.defaults import patterns, url            
         
         urlpatterns = super(TranslatableAdmin, self).get_urls()
 
@@ -136,7 +162,10 @@ class TranslatableAdmin(ModelAdmin, TranslatableModelAdminMixin):
                 return self.admin_site.admin_view(view)(*args, **kwargs)
             return update_wrapper(wrapper, view)
 
-        info = self.model._meta.app_label, self.model._meta.module_name
+        if django.VERSION >= (1, 6):
+            info = self.model._meta.app_label, self.model._meta.model_name
+        else:
+            info = self.model._meta.app_label, self.model._meta.module_name
 
         urlpatterns = patterns('',
             url(r'^(.+)/delete-translation/(.+)/$',
@@ -151,10 +180,20 @@ class TranslatableAdmin(ModelAdmin, TranslatableModelAdminMixin):
         add_view and change_view.
         """
         
-        if self.declared_fieldsets:
-            fields = flatten_fieldsets(self.declared_fieldsets)
+        if django.VERSION >= (1, 6):
+            # From v1.6 on, using get_fieldsets is ok, as long as no 'fields'
+            # argument was given. It allows dynamic fieldsets on admin form.
+            if 'fields' in kwargs:
+                fields = kwargs.pop('fields')
+            else:
+                fields = flatten_fieldsets(self.get_fieldsets(request, obj))
         else:
-            fields = None
+            # On previous versions, calling get_fieldsets triggers infinite recursion
+            # and we should stick to statically declared fieldsets
+            if self.declared_fieldsets:
+                fields = flatten_fieldsets(self.declared_fieldsets)
+            else:
+                fields = None
         if self.exclude is None:
             exclude = []
         else:
@@ -196,15 +235,18 @@ class TranslatableAdmin(ModelAdmin, TranslatableModelAdminMixin):
     def response_change(self, request, obj):
         redirect = super(TranslatableAdmin, self).response_change(request, obj)
         uri = iri_to_uri(request.path)
-        app_label, model_name = self.model._meta.app_label, self.model._meta.module_name
-        if redirect['Location'] in (uri, "../add/", reverse('admin:%s_%s_add' % (app_label, model_name))):
+        if django.VERSION >= (1, 6):
+            app_label, model_name = self.model._meta.app_label, self.model._meta.model_name
+        else:
+            app_label, model_name = self.model._meta.app_label, self.model._meta.module_name
+        if redirect['Location'] in (uri, "../add/", self.reverse('admin:%s_%s_add' % (app_label, model_name))):
             if self.query_language_key in request.GET:
                 redirect['Location'] = '%s?%s=%s' % (redirect['Location'],
                     self.query_language_key, request.GET[self.query_language_key])
         return redirect
     
     @csrf_protect_m
-    @transaction.commit_on_success
+    @atomic
     def delete_translation(self, request, object_id, language_code):
         "The 'delete translation' admin view for this model."
         opts = self.model._meta
@@ -256,8 +298,9 @@ class TranslatableAdmin(ModelAdmin, TranslatableModelAdminMixin):
             )
 
             if not self.has_change_permission(request, None):
-                return HttpResponseRedirect(reverse('admin:index'))
-            return HttpResponseRedirect(reverse('admin:%s_%s_changelist' % (opts.app_label, opts.module_name)))
+                return HttpResponseRedirect(self.reverse('admin:index'))
+            model_name = opts.model_name if django.VERSION >= (1, 6) else opts.module_name
+            return HttpResponseRedirect(self.reverse('admin:%s_%s_changelist' % (opts.app_label, model_name)))
 
         object_name = '%s Translation' % force_unicode(opts.verbose_name)
 
@@ -306,22 +349,23 @@ class TranslatableAdmin(ModelAdmin, TranslatableModelAdminMixin):
     
     def get_object(self, request, object_id):
         obj = super(TranslatableAdmin, self).get_object(request, object_id)
-        if obj:
-            return obj
-        queryset = self.model.objects.untranslated()
-        model = self.model
-        try:
-            object_id = model._meta.pk.to_python(object_id)
-            obj = queryset.get(pk=object_id)
-        except (model.DoesNotExist, ValidationError):
+        if obj is None: # object was not in queryset, bail out
             return None
-        new_translation = model._meta.translations_model()
-        new_translation.language_code = self._language(request)
-        new_translation.master = obj
-        setattr(obj, model._meta.translations_cache, new_translation)
+
+        # object was in queryset - need to make sure we got the right translation
+        # we use getattr to trigger a load if instance exists but translation was
+        # not cached yet. Should not happen with current code, but is correct,
+        # future-proof behavior.
+        language_code = getattr(obj, 'language_code', None)
+        request_lang = self._language(request)
+        if language_code is None or language_code != request_lang:
+            # if language does not match that of request, we know request_lang
+            # does not exist, because it was the first language in the use_fallbacks
+            # list. We prepare it as a new translation.
+            obj.translate(request_lang)
         return obj
 
-    def queryset(self, request):
+    def get_queryset(self, request):
         language = self._language(request)
         languages = [language,]
         for lang in FALLBACK_LANGUAGES:
@@ -333,6 +377,8 @@ class TranslatableAdmin(ModelAdmin, TranslatableModelAdminMixin):
         if ordering:
             qs = qs.order_by(*ordering)
         return qs
+    if django.VERSION < (1, 8):
+        queryset = get_queryset
 
     def get_change_form_base_template(self):
         opts = self.model._meta
@@ -360,10 +406,20 @@ class TranslatableInlineModelAdmin(InlineModelAdmin, TranslatableModelAdminMixin
 
     def get_formset(self, request, obj=None, **kwargs):
         """Returns a BaseInlineFormSet class for use in admin add/change views."""
-        if self.declared_fieldsets:
-            fields = flatten_fieldsets(self.declared_fieldsets)
+        if django.VERSION >= (1, 6):
+            # From v1.6 on, using get_fieldsets is ok, as long as no 'fields'
+            # argument was given. It allows dynamic fieldsets on admin form.
+            if 'fields' in kwargs:
+                fields = kwargs.pop('fields')
+            else:
+                fields = flatten_fieldsets(self.get_fieldsets(request, obj))
         else:
-            fields = None
+            # On previous versions, calling get_fieldsets triggers infinite recursion
+            # and we should stick to statically declared fieldsets
+            if self.declared_fieldsets:
+                fields = flatten_fieldsets(self.declared_fieldsets)
+            else:
+                fields = None
         if self.exclude is None:
             exclude = []
         else:
@@ -374,7 +430,7 @@ class TranslatableInlineModelAdmin(InlineModelAdmin, TranslatableModelAdminMixin
         # default
         exclude = exclude or None
         defaults = {
-            "form": self.get_form(request, obj),
+            "form": self.get_form(request, obj, fields=fields),
             #"formset": self.formset,
             "fk_name": self.fk_name,
             "fields": fields,
@@ -412,10 +468,20 @@ class TranslatableInlineModelAdmin(InlineModelAdmin, TranslatableModelAdminMixin
         Returns a Form class for use in the admin add view. This is used by
         add_view and change_view.
         """
-        if self.declared_fieldsets:
-            fields = flatten_fieldsets(self.declared_fieldsets)
+        if django.VERSION >= (1, 6):
+            # From v1.6 on, using get_fieldsets is ok, as long as no 'fields'
+            # argument was given. It allows dynamic fieldsets on admin form.
+            if 'fields' in kwargs:
+                fields = kwargs.pop('fields')
+            else:
+                fields = flatten_fieldsets(self.get_fieldsets(request, obj))
         else:
-            fields = None
+            # On previous versions, calling get_fieldsets triggers infinite recursion
+            # and we should stick to statically declared fieldsets
+            if self.declared_fieldsets:
+                fields = flatten_fieldsets(self.declared_fieldsets)
+            else:
+                fields = None
         if self.exclude is None:
             exclude = []
         else:
@@ -449,7 +515,7 @@ class TranslatableInlineModelAdmin(InlineModelAdmin, TranslatableModelAdminMixin
     """
 # Should be added
     @csrf_protect_m
-    @transaction.commit_on_success
+    @atomic
     def delete_translation(self, request, object_id, language_code):
         "The 'delete translation' admin view for this model."
         opts = self.model._meta
@@ -546,7 +612,7 @@ class TranslatableInlineModelAdmin(InlineModelAdmin, TranslatableModelAdminMixin
     def delete_model_translation(self, request, obj):
         obj.delete()
     """
-    def queryset(self, request):
+    def get_queryset(self, request):
         language = self._language(request)
         qs = self.model._default_manager.all()#.language(language)
         # TODO: this should be handled by some parameter to the ChangeList.
@@ -554,7 +620,8 @@ class TranslatableInlineModelAdmin(InlineModelAdmin, TranslatableModelAdminMixin
         if ordering:
             qs = qs.order_by(*ordering)
         return qs
-
+    if django.VERSION < (1, 8):
+        queryset = get_queryset
 
 class TranslatableStackedInline(TranslatableInlineModelAdmin):
     template = 'admin/hvad/edit_inline/stacked.html'
