@@ -1,13 +1,16 @@
 from collections import defaultdict
+import django
 from django.conf import settings
 from django.db import models, transaction, IntegrityError
-from django.db.models.query import (QuerySet, ValuesQuerySet, DateQuerySet, 
-    CHUNK_SIZE)
+from django.db.models.query import QuerySet, ValuesQuerySet, DateQuerySet
+try:
+    from django.db.models.query import CHUNK_SIZE
+except ImportError:
+    CHUNK_SIZE = 100
 from django.db.models.query_utils import Q
 from django.utils.translation import get_language
 from hvad.fieldtranslator import translate
 from hvad.utils import combine
-import django
 import logging
 import sys
 
@@ -92,14 +95,14 @@ class TranslationQueryset(QuerySet):
         DateQuerySet: DatesMixin,
     }
     
-    def __init__(self, model=None, query=None, using=None, real=None):
+    def __init__(self, *args, **kwargs):
         self._local_field_names = None
         self._field_translator = None
-        self._real_manager = real
+        self._real_manager = kwargs.pop('real', None)
         self._fallback_manager = None
         self._language_code = None
         self._forced_unique_fields = []  # Used for select_related
-        super(TranslationQueryset, self).__init__(model=model, query=query, using=using)
+        super(TranslationQueryset, self).__init__(*args, **kwargs)
 
         # After super(), make sure we retrieve the shared model:
         if not self.query.select_related:
@@ -177,7 +180,7 @@ class TranslationQueryset(QuerySet):
         for child in q.children:
             if isinstance(child, Q):
                 newq = self._recurse_q(child)
-                newchildren.append(self._recurse_q(newq))
+                newchildren.append(newq)
             else:
                 key, value = child
                 newchildren.append((self.field_translator.get(key), value))
@@ -219,7 +222,7 @@ class TranslationQueryset(QuerySet):
                 return type(value.__name__, (value, klass, TranslationQueryset,), {})
         return klass
     
-    def _get_shared_query_set(self):
+    def _get_shared_queryset(self):
         qs = super(TranslationQueryset, self)._clone()
         qs.__class__ = QuerySet
         # un-select-related the 'master' relation
@@ -227,6 +230,24 @@ class TranslationQueryset(QuerySet):
         accessor = self.shared_model._meta.translations_accessor
         # update using the real manager
         return self._real_manager.filter(**{'%s__in' % accessor:qs})
+
+    def _scan_for_language_where_node(self, children):
+        found = False
+        for node in children:
+            try:
+                if django.VERSION >= (1, 7):
+                    field_name = node.lhs.target.name
+                else:
+                    field_name = node[0].field.name
+            except (TypeError, AttributeError):
+                if node.children:
+                    found = self._scan_for_language_where_node(node.children)
+            else:
+                found = field_name == 'language_code'
+
+            if found:
+                # No need to continue
+                return True
     
     #===========================================================================
     # Queryset/Manager API 
@@ -288,14 +309,7 @@ class TranslationQueryset(QuerySet):
                 qs = self.language(language_code)
                 found = True
         else:
-            for where in qs.query.where.children:
-                if where.children:
-                    for child in where.children:
-                        if child[0].field.name == 'language_code':
-                            found = True
-                            break
-                if found:
-                    break
+            found = self._scan_for_language_where_node(qs.query.where.children)
         if not found:
             qs = self.language()
         # self.iterator already combines! Isn't that nice?
@@ -333,14 +347,18 @@ class TranslationQueryset(QuerySet):
                 obj.save(force_insert=True, using=self.db)
                 transaction.savepoint_commit(sid, using=self.db)
                 return obj, True
-            except IntegrityError, e:
+            except IntegrityError:
                 transaction.savepoint_rollback(sid, using=self.db)
                 exc_info = sys.exc_info()
                 try:
                     return self.get(**lookup), False
                 except self.model.DoesNotExist:
                     # Re-raise the IntegrityError with its original traceback.
-                    raise exc_info[1], None, exc_info[2]
+                    raise exc_info[1]
+
+    if django.VERSION >= (1, 7):
+        def update_or_create(self, defaults=None, **kwargs):
+            raise NotImplementedError()
 
     def filter(self, *args, **kwargs):
         newargs, newkwargs = self._translate_args_kwargs(*args, **kwargs)
@@ -367,16 +385,20 @@ class TranslationQueryset(QuerySet):
         return super(TranslationQueryset, self).latest(field_name)
 
     def in_bulk(self, id_list):
-        raise NotImplementedError()
+        if not id_list:
+            return {}
+        qs = self.filter(pk__in=id_list)
+        qs.query.clear_ordering(force_empty=True)
+        return dict([(obj._get_pk_val(), obj) for obj in qs.iterator()])
 
     def delete(self):
-        qs = self._get_shared_query_set()
+        qs = self._get_shared_queryset()
         qs.delete()
     delete.alters_data = True
     
     def delete_translations(self):
         self.update(master=None)
-        super(TranslationQueryset, self).delete()
+        self.model.objects.filter(master__isnull=True).delete()
     delete_translations.alters_data = True
         
     def update(self, **kwargs):
@@ -385,7 +407,7 @@ class TranslationQueryset(QuerySet):
         if translated:
             count += super(TranslationQueryset, self).update(**translated)
         if shared:
-            shared_qs = self._get_shared_query_set()
+            shared_qs = self._get_shared_queryset()
             count += shared_qs.update(**shared)
         return count
     update.alters_data = True
@@ -400,8 +422,8 @@ class TranslationQueryset(QuerySet):
 
     def dates(self, field_name, kind=None, order='ASC'):
         field_name = self.field_translator.get(field_name)
-        if int(django.get_version().split('.')[1][0]) <= 2:
-            from nani.compat.date import DateQuerySet
+        if django.VERSION <= (1, 2):
+            from hvad.compat.date import DateQuerySet
             return self._clone(klass=DateQuerySet, setup=True,
                 _field_name=field_name, _kind=kind, _order=order)
         return super(TranslationQueryset, self).dates(field_name, kind=kind, order=order)
@@ -429,8 +451,14 @@ class TranslationQueryset(QuerySet):
                 query_key = 'master__%s' % query_key
             except models.FieldDoesNotExist:
                 field, model, direct, _ = self.model._meta.get_field_by_name(bits[0])
-            if not model:
-                model = field.rel.to if direct else field.model
+
+            if direct:  # field is on model
+                if field.rel:    # field is a foreign key, follow it
+                    model = field.rel.to
+                else:            # field is a regular field
+                    raise AssertionError('Cannot select_related: %s is a regular field' % query_key)
+            else:       # field is a m2m or reverse fk, follow it
+                model = field.model
 
             if hasattr(model._meta, 'translations_accessor'):  # if issubclass(model, TranslatableModel):
                 # This is a relation to a translated model,
@@ -440,8 +468,13 @@ class TranslationQueryset(QuerySet):
                 related_model_keys.append(query_key)  # Select the related model
                 related_model_keys.append('%s__%s' % (query_key, model._meta.translations_accessor))  # and its translation model
 
-                # We need to force this to be a LEFT OUTER join, so we explicitly add the join:
-                related_model_explicit_joins.append((field.model._meta.db_table, model._meta.db_table, bits[0] + "_id", 'id'))
+                # We need to force this to be a LEFT OUTER join, so we explicitly add the join.
+                # Django 1.6 changes the footprint of the Query.join method. See https://code.djangoproject.com/ticket/19385
+                if django.VERSION < (1, 6):
+                    join_data = (field.model._meta.db_table, model._meta.db_table, bits[0] + "_id", 'id')
+                else:
+                    join_data = (field, (field.model._meta.db_table, model._meta.db_table, ((bits[0] + "_id", 'id'),)))
+                related_model_explicit_joins.append(join_data)
                 # And we are going to force the query to treat the language join as one-to-one,
                 # so we need to filter for the desired language:
                 related_model_extra_filters.append(('%s__%s__language_code' % (query_key, model._meta.translations_accessor), self._language_code))
@@ -457,7 +490,12 @@ class TranslationQueryset(QuerySet):
         obj = self._clone()
         obj.query.get_compiler(obj.db).fill_related_selections()  # seems to be necessary; not sure why
         for j in related_model_explicit_joins:
-            obj.query.join(j, outer_if_first=True)
+            if django.VERSION >= (1, 6):
+                kwargs = {'join_field': j[0]}
+                j = j[1]
+            else:
+                kwargs = {}
+            obj.query.join(j, nullable=True, **kwargs)
         for f in related_model_extra_filters:
             f1 = {f[0]: f[1]}
             f2 = {f[0]: None}  # Allow select_related() to fetch objects with a relation set to NULL
@@ -582,10 +620,12 @@ class TranslationManager(models.Manager):
     # API 
     #===========================================================================
 
+    queryset_class = TranslationQueryset
+
     def using_translations(self):
         if not hasattr(self, '_real_manager'):
             self.contribute_real_manager()
-        qs = TranslationQueryset(self.translations_model, using=self.db, real=self._real_manager)
+        qs = self.queryset_class(self.translations_model, using=self.db, real=self._real_manager)
         if hasattr(self, 'core_filters'):
             qs = qs._next_is_sticky().filter(**(self.core_filters))
         return qs
@@ -594,7 +634,10 @@ class TranslationManager(models.Manager):
         return self.using_translations().language(language_code)
 
     def untranslated(self):
-        return self._fallback_manager.get_query_set()
+        if django.VERSION >= (1, 6):
+            return self._fallback_manager.get_queryset()
+        else:
+            return self._fallback_manager.get_query_set()
 
     #===========================================================================
     # Internals
@@ -607,7 +650,7 @@ class TranslationManager(models.Manager):
         """
         return self.model._meta.translations_model
 
-    #def get_query_set(self):
+    #def get_queryset(self):
     #    """
     #    Make sure that querysets inherit the methods on this manager (chaining)
     #    """
@@ -646,7 +689,7 @@ class FallbackQueryset(QuerySet):
         """
         The logic for this method was taken from django-polymorphic by Bert
         Constantin (https://github.com/bconstantin/django_polymorphic) and was
-        slightly altered to fit the needs of django-nani.
+        slightly altered to fit the needs of django-hvad.
         """
         # get the primary keys of the shared model results
         base_ids = [obj.pk for obj in base_results]
@@ -684,7 +727,7 @@ class FallbackQueryset(QuerySet):
         """
         The logic for this method was taken from django-polymorphic by Bert
         Constantin (https://github.com/bconstantin/django_polymorphic) and was
-        slightly altered to fit the needs of django-nani.
+        slightly altered to fit the needs of django-hvad.
         """
         base_iter = super(FallbackQueryset, self).iterator()
 
@@ -697,7 +740,7 @@ class FallbackQueryset(QuerySet):
                 # get the next "chunk" of results
                 for i in range(CHUNK_SIZE):
                     try:
-                        instance = base_iter.next()
+                        instance = next(base_iter)
                         base_result_objects.append(instance)
                     except StopIteration:
                         reached_end = True
@@ -739,11 +782,15 @@ class TranslationFallbackManager(models.Manager):
     using `use_fallbacks()` to enable per object language fallback.
     """
     def use_fallbacks(self, *fallbacks):
-        return self.get_query_set().use_fallbacks(*fallbacks)
-    
-    def get_query_set(self):
+        if django.VERSION >= (1, 6):
+            return self.get_queryset().use_fallbacks(*fallbacks)
+        else:
+            return self.get_query_set().use_fallbacks(*fallbacks)
+
+    def get_queryset(self):
         qs = FallbackQueryset(self.model, using=self.db)
         return qs
+    get_query_set = get_queryset        # old name for Django < 1.6
 
 
 #===============================================================================
@@ -783,15 +830,15 @@ class TranslationAwareQueryset(QuerySet):
         language_joins = []
         for child in q.children:
             if isinstance(child, Q):
-                newq = self._recurse_q(child)
-                newchildren.append(self._recurse_q(newq))
+                newq, langjoins = self._recurse_q(child)
+                newchildren.append(newq)
             else:
                 key, value = child
                 newkey, langjoins =translate(key, self.model)
                 newchildren.append((newkey, value))
-                for langjoin in langjoins:
-                    if langjoin not in language_joins:
-                        language_joins.append(langjoin)
+            for langjoin in langjoins:
+                if langjoin not in language_joins:
+                    language_joins.append(langjoin)
         q.children = newchildren
         return q, language_joins
     
@@ -838,7 +885,11 @@ class TranslationAwareQueryset(QuerySet):
         return self._filter_extra(extra_filters).latest(field_name)
 
     def in_bulk(self, id_list):
-        raise NotImplementedError()
+        if not id_list:
+            return {}
+        qs = self.filter(pk__in=id_list)
+        qs.query.clear_ordering(force_empty=True)
+        return dict([(obj._get_pk_val(), obj) for obj in qs.iterator()])
 
     def values(self, *fields):
         fields, extra_filters = self._translate_fieldnames(fields)
@@ -897,11 +948,15 @@ class TranslationAwareQueryset(QuerySet):
 
 class TranslationAwareManager(models.Manager):
     def language(self, language_code=None):
-        return self.get_query_set().language(language_code)
-        
-    def get_query_set(self):
+        if django.VERSION >= (1, 6):
+            return self.get_queryset().language(language_code)
+        else:
+            return self.get_query_set().language(language_code)
+
+    def get_queryset(self):
         qs = TranslationAwareQueryset(self.model, using=self.db)
         return qs
+    get_query_set = get_queryset        # old name for Django < 1.6
 
 
 #===============================================================================
