@@ -4,12 +4,15 @@ from django.db import models
 from django.db.models.base import ModelBase
 from django.db.models.signals import post_save
 from django.utils.translation import get_language
+from hvad.compat.metaclasses import with_metaclass
 from hvad.descriptors import LanguageCodeAttribute, TranslatedAttribute
 from hvad.manager import TranslationManager, TranslationsModelManager
 from hvad.utils import SmartGetFieldByName
-from types import MethodType
+from hvad.compat.method_type import MethodType
 import sys
 
+# maybe there should be an extra settings for this
+FALLBACK_LANGUAGES = tuple( code for code, name in settings.LANGUAGES )
 
 def create_translations_model(model, related_name, meta, **fields):
     """
@@ -104,7 +107,7 @@ class TranslatableModelBase(ModelBase):
             return super_new(cls, name, bases, attrs)
         new_model = super_new(cls, name, bases, attrs)
         opts = new_model._meta
-        if not opts.abstract and not isinstance(new_model.objects, TranslationManager):
+        if not opts.abstract and not isinstance(new_model._default_manager, TranslationManager):
             raise ImproperlyConfigured(
                 "The default manager on a TranslatableModel must be a "
                 "TranslationManager instance or an instance of a subclass of "
@@ -126,7 +129,7 @@ class TranslatableModelBase(ModelBase):
                     concrete_model = concrete_model._meta.proxy_for_model
 
         found = False
-        for relation in concrete_model.__dict__.keys():
+        for relation in list(concrete_model.__dict__.keys()):
             try:
                 obj = getattr(new_model, relation)
             except AttributeError:
@@ -156,8 +159,7 @@ class TranslatableModelBase(ModelBase):
         
         if not isinstance(opts.get_field_by_name, SmartGetFieldByName):
             smart_get_field_by_name = SmartGetFieldByName(opts.get_field_by_name)
-            opts.get_field_by_name = MethodType(smart_get_field_by_name , opts,
-                                                opts.__class__)
+            opts.get_field_by_name = MethodType(smart_get_field_by_name , opts)
         
         return new_model
     
@@ -165,12 +167,11 @@ class TranslatableModelBase(ModelBase):
 class NoTranslation(object):
     pass
 
-class TranslatableModel(models.Model):
+
+class TranslatableModel(with_metaclass(TranslatableModelBase, models.Model)):
     """
     Base model for all models supporting translated fields (via TranslatedFields).
     """
-    __metaclass__ = TranslatableModelBase
-    
     # change the default manager to the translation manager
     objects = TranslationManager()
     
@@ -189,7 +190,7 @@ class TranslatableModel(models.Model):
         
         # filter out all the translated fields (including 'master' and 'language_code')
         primary_key_names = ('pk', self._meta.pk.name)
-        for key in kwargs.keys():
+        for key in list(kwargs.keys()):
             if key in self._translated_field_names:
                 if not key in primary_key_names:
                     # we exclude the pk of the shared model
@@ -203,7 +204,7 @@ class TranslatableModel(models.Model):
         # there was at least one of the translated fields (or a language_code) 
         # in kwargs. We need to do magic.
         # extract all the shared fields (including the pk)
-        for key in kwargs.keys():
+        for key in list(kwargs.keys()):
             if key in self._shared_field_names:
                 skwargs[key] = kwargs.pop(key)
         # do the regular init minus the translated fields
@@ -287,40 +288,34 @@ class TranslatableModel(models.Model):
             return stuff
 
         # get all translations
-        translations = self._meta.translations_model.objects.filter(master__pk=self.pk)
+        translations = getattr(self, self._meta.translations_accessor).all()
 
-        # make them a nice dict
-        translation_dict = {}
-        for translation in translations:
-            translation_dict[translation.language_code] = translation
-
-        # see if we have the right language
-        correct_language = translation_dict.get(get_language())
-        if correct_language:
-            # CACHE IT!
-            setattr(self, self._meta.translations_cache, correct_language)
-            return getattr(correct_language, name, default)
-
-        # Nope? Ok, let's try the default language for the installation
-        default_language = translation_dict.get(settings.LANGUAGE_CODE)
-        if default_language:
-            # CACHE IT!
-            setattr(self, self._meta.translations_cache, default_language)
-            return getattr(default_language, name, default)
-
-        # *Sigh*, OK, any available langauge?
-        try:
-            any_language = translation_dict.values()[0]
-        except IndexError:
-            # OK, can't say we didn't try!
+        # if no translation exists, bail out now
+        if len(translations) == 0:
             return default
 
-        setattr(self, self._meta.translations_cache, any_language)
-        return getattr(any_language, name, default)
+        # organize translations into a nice dict
+        translation_dict = dict((t.language_code, t) for t in translations)
+
+        # see if we have the right language, or any language in fallbacks
+        for code in (get_language(), settings.LANGUAGE_CODE) + FALLBACK_LANGUAGES:
+            try:
+                translation = translation_dict[code]
+            except KeyError:
+                continue
+            break
+        else:
+            # none of the fallbacks was found, pick an arbitrary translation
+            translation = translation_dict.popitem()[1]
+
+        setattr(self, self._meta.translations_cache, translation)
+        return getattr(translation, name, default)
 
     def get_available_languages(self):
-        manager = self._meta.translations_model.objects
-        return manager.filter(master=self).values_list('language_code', flat=True)
+        qs = getattr(self, self._meta.translations_accessor).all()
+        if qs._result_cache is not None:
+            return [obj.language_code for obj in qs]
+        return qs.values_list('language_code', flat=True)
     
     #===========================================================================
     # Internals
